/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "rollball.hpp"
#include "fggl/data/model.hpp"
#include "fggl/data/procedural.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/input/camera_input.h"

struct Prefabs {
	fggl::ecs3::entity_t wallX;
	fggl::ecs3::entity_t wallZ;
	fggl::ecs3::entity_t floor;
	fggl::ecs3::entity_t collectable;
	fggl::ecs3::entity_t player;
};

static void setupPrefabs(fggl::ecs3::World& world, Prefabs& prefabs) {

	{
		// walls
		prefabs.wallX = world.create(true);
		world.add<fggl::math::Transform>(prefabs.wallX);

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";

		// create a procedural cube, but scale the model to be a long wall
		auto transform = glm::scale(fggl::data::OFFSET_NONE, {1.0f, 5.f, 41.0f});
		fggl::data::make_cube(meshComponent.mesh, transform );
		world.set<fggl::data::StaticMesh>(prefabs.wallX, &meshComponent);
	}

	{
		// walls
		prefabs.wallZ = world.create(true);
		world.add<fggl::math::Transform>(prefabs.wallZ);

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";

		// create a procedural cube, but scale the model to be a long wall
		auto transform = glm::scale(fggl::data::OFFSET_NONE, {39.f, 5.f, 1.0f});
		fggl::data::make_cube(meshComponent.mesh, transform );
		world.set<fggl::data::StaticMesh>(prefabs.wallZ, &meshComponent);
	}

	{
		// floor
		prefabs.floor = world.create(true);
		world.add<fggl::math::Transform>(prefabs.floor);

		auto* rb = world.add<fggl::phys::RigidBody>(prefabs.floor);
		rb->mass = fggl::phys::STATIC_MASS;

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";

		// create a procedural cube, but scale the model to be a long wall
		auto transform = glm::scale(fggl::data::OFFSET_NONE, {40.f, 0.5f, 40.0f});

		// TODO proper planes
		fggl::data::make_cube(meshComponent.mesh, transform );
		world.set<fggl::data::StaticMesh>(prefabs.floor, &meshComponent);
	}

	{
		// player (cube because my sphere function doesn't exist yet
		prefabs.player = world.create(true);
		world.add<fggl::math::Transform>(prefabs.player);
		world.add<fggl::phys::RigidBody>(prefabs.player);

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";
		fggl::data::make_sphere(meshComponent.mesh);
		world.set<fggl::data::StaticMesh>(prefabs.player, &meshComponent);
	}

	{
		// collectable
		prefabs.collectable = world.create(true);
		world.add<fggl::math::Transform>(prefabs.collectable);

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";
		fggl::data::make_cube(meshComponent.mesh);
		world.set<fggl::data::StaticMesh>(prefabs.collectable, &meshComponent);
	}

}

static void setupCamera(fggl::ecs3::World& world) {
	auto prototype = world.create(false);

	// setup camera position/transform
	auto* transform = world.add<fggl::math::Transform>(prototype);
	if ( transform != nullptr) {
		transform->origin(glm::vec3(10.0f, 3.0f, 10.0f));
	}

	// setup camera components
	world.add<fggl::gfx::Camera>(prototype);
	auto* cameraKeys = world.add<fggl::input::FreeCamKeys>(prototype);
	if ( cameraKeys != nullptr ) {
		cameraKeys->forward = glfwGetKeyScancode(GLFW_KEY_W);
		cameraKeys->backward = glfwGetKeyScancode(GLFW_KEY_S);
		cameraKeys->left = glfwGetKeyScancode(GLFW_KEY_A);
		cameraKeys->right = glfwGetKeyScancode(GLFW_KEY_D);
		cameraKeys->rotate_cw = glfwGetKeyScancode(GLFW_KEY_Q);
		cameraKeys->rotate_ccw = glfwGetKeyScancode(GLFW_KEY_E);
	}
}

static void setupEnvironment(fggl::ecs3::World& world, const Prefabs prefabs, fggl::math::vec2& size) {
	{
		auto northWall = world.copy(prefabs.wallX);
		auto* transform = world.get<fggl::math::Transform>(northWall);
		transform->origin({size.x/2, 0.0f, 0.0f});
	}

	{
		auto southWall = world.copy(prefabs.wallX);
		auto* transform = world.get<fggl::math::Transform>(southWall);
		transform->origin({-size.x/2, 0.0f, 0.0f});
	}

	{
		auto westWall = world.copy(prefabs.wallZ);
		auto* transform = world.get<fggl::math::Transform>(westWall);
		transform->origin({0.0f, 0.0f, -size.y/2});
	}

	{
		auto eastWall = world.copy(prefabs.wallZ);
		auto* transform = world.get<fggl::math::Transform>(eastWall);
		transform->origin({0.0f, 0.0f, size.y/2});
	}

	{
		auto floor = world.copy(prefabs.floor);
		auto *transform = world.get<fggl::math::Transform>(floor);
		transform->origin({0.0f, -2.5f, 0.0f});
	}

	{
		// player just starts off as the prefab dictates
		auto player = world.copy(prefabs.player);
	}

	{
		// collectables
		std::array<fggl::math::vec3, 3> collectPos {{
			{-5.0f, 0.0f, 12.0f},
			{15.0f, 0.0f, 0.5f},
			{6.0f, 0.0f, -15.0f}
		}};

		// build the collectables
		for (auto& pos : collectPos) {
			auto collectable = world.copy(prefabs.collectable);
			auto* transform = world.get<fggl::math::Transform>(collectable);
			transform->origin(pos);
		}
	}
}

enum camera_type { cam_free, cam_arcball };
static camera_type cam_mode = cam_free;

static void process_camera(fggl::ecs3::World& ecs, const fggl::input::Input& input) {
	auto cameras = ecs.findMatching<fggl::gfx::Camera>();
	fggl::ecs3::entity_t cam = cameras[0];

	auto camTransform = ecs.get<fggl::math::Transform>(cam);
	auto camComp = ecs.get<fggl::gfx::Camera>(cam);

	const glm::vec3 dir = ( camTransform->origin() - camComp->target );
	const glm::vec3 forward = glm::normalize( dir );

	// scroll wheel
	glm::vec3 motion(0.0f);
	float delta = input.mouse.axis( fggl::input::MouseAxis::SCROLL_Y );
	if ( (glm::length( dir ) < 25.0f && delta < 0.0f) || (glm::length( dir ) > 2.5f && delta > 0.0f) )
		motion -= (forward * delta);
	camTransform->origin( camTransform->origin() + motion );

	if ( cam_mode == cam_arcball || input.mouse.down( fggl::input::MouseButton::MIDDLE ) ) {
		fggl::input::process_arcball(ecs, input, cam);
	} else if ( cam_mode == cam_free ) {
		fggl::input::process_freecam(ecs, input, cam);
	}
	fggl::input::process_edgescroll( ecs, input, cam );
}

namespace demo {

	RollBall::RollBall(fggl::App &app) : Game(app) {
	}

	void RollBall::activate() {
		Game::activate();

		Prefabs prefabs{};
		setupPrefabs(world(), prefabs);

		// actual scene objects
		setupCamera(world());

		// create a 20x20 grid
		fggl::math::vec2 size{40.0f, 40.0f};
		setupEnvironment(world(), prefabs, size);
	}

	void RollBall::update() {
		Game::update();
		process_camera(world(), input());
	}

	void RollBall::render(fggl::gfx::Graphics &gfx) {
		Game::render(gfx);
	}

} // namespace demo

